/*!
 * @file mapped_water.h
 *
*/

#ifndef _MAPPED_WATER_H
#define _MAPPED_WATER_H

/* INCLUDES */
#include "world_entity.h"
#include "material.h"
#include "shader.h"

class MappedWater : public WorldEntity
{
  public:
    MappedWater(const TiXmlElement* root = NULL);
    virtual ~MappedWater();

    void loadParams(const TiXmlElement* root);


    void activateReflection();
    void deactivateReflection();

    void activateRefraction();
    void deactivateRefraction();

    void draw() const;
    void tick(float dt);

  private:
    void setLightPosition(float x, float y, float z) { this->lightPos = Vector(x,y,z); };
    void setHeight(float height);

  private:
    float               waterHeight;                //!< y-coord of the Water

    float               move;                       //!< textures coords, speeds, positions for the shaded textures....
    float               move2;                      //!<
    float               g_WaterUV;                  //!<
    float               g_WaterFlow;                //!<
    float               refrUV;                     //!<
    float               normalUV;                   //!<
    float               kNormalMapScale;            //!<

    int                 textureSize;                //!< size of the texture
    Vector              lightPos;
    Material            mat;
    Shader*             shader;
    Shader::Uniform*    cam_uni;                        //!< uniform that is used for the camera position

};

#endif  /* _MAPPED_WATER_H */
